/*
 * JFCP43 - 20190312
 * Ce programme permet de commander jusqu'à 5 aiguilles et 5 itineraires
 * avec affichage associe de leds sur le TCO
 * 
 * Le passage du mode 'aiguilles' au mode 'itineraires' se fait via l'interrupteur MODE_btn
 * et sa led associee MODE_led
 * Cet interrupteur bipolaire gere aussi le raccordement des boutons de commandes au neutre
 * permettant d'utiliser les memes entrées de l'Arduino pour les boutons poussoirs des aiguilles
 * et ceux des itineraires en n'activant que ceux du mode actif
 * 
 * Les aiguilles sont commandees par des paires de sorties a partir de la broche FIRST_aig :
 * Aiguille 1 : bobine 1 = AIG_first, bobine 2 = AIG_first + 1, etc...
 * Ces sorties pilotent directement des ULN2803 envoyant l'impulsion aux aiguilles du reseau
 * dont la duree, en ms est modifiable via ACTIVATION
 * 
 * La commutation des aiguilles se fait via des boutons poussoirs
 * traites avec anti-rebond logiciel reglable via DELAI_ANTI_REBOND
 * 
 * Les leds reportant l'etat des aiguilles ou des itineraires sur le TCO sont des
 * leds WS2812b => adressables une à une avec gestion de leur couleur et de leur luminosite
 * 
 */

/***************************************************************************
 * 
 * PARAMETRES GENERAUX
 * 
 ***************************************************************************/

// Récupération de la bibliothèque commandant les leds WS2812b
#include <Adafruit_NeoPixel.h>

#define ACTIVATION        200 // duree d'activation des bobines en ms
#define DELAI_ANTI_REBOND 100  // pour la gestion logicielle du rebond des interrupteurs

const byte MODE_btn = 14;  // broche A0 : interrupteur de choix entre mode aiguilles (HIGH) et itineraires (LOW)
const byte MODE_led = 13;  // broche D13 : led associee
boolean MODE_etat = LOW;  // dernier etat associe, mode 'aiguilles' par defaut


/***************************************************************************
 * 
 * AIGUILLES
 * 
 ***************************************************************************/
const byte AIG_nb = 5;      // nombre d'aiguilles exploitees (5 maxi)
const byte AIG_first = 2;   // broche de depart des aiguilles
byte AIG_broche;            // broche a piloter pour l'aiguille en cours de traitement
boolean AIG_etat[AIG_nb];   // etat des aiguilles

// etat par defaut des aiguilles : directe = 0, deviee = 1
const boolean AIG_defaut[AIG_nb] = {0, 1, 1, 0, 0};

// leds associees à chaque aiguille : directe puis deviee
const byte AIG_led[AIG_nb][2] = {
  {3, 1},     // aiguille 1 : directe = led 3, deviee = led 1
  {6, 5},     // aiguille 2 : directe = led 6, deviee = led 5
  {10, 9},    // aiguille 3
  {8, 7},     // aiguille 4
  {14, 13}    // aiguille 5
};

/***************************************************************************
 * 
 * ITINERAIRES
 * 
 ***************************************************************************/
const byte ITI_nb = 4;        // nombre d'itineraires exploites (5 maxi)
const byte ITI_led_max = 6;   // nb de leds maxi à allumer par itineraire (maxi NB_leds du ruban WS2812b)
const byte ITI_defaut = 1;    // Itineraire a activer par defaut

// leds a allumer par itineraire :
// definir ici le n° des leds a allumer pour chaque itineraire, dans leur ordre d'allumage
const byte ITI_led[ITI_nb][ITI_led_max] = {
  {3, 5, 4, 14, 12, 0},
  {3, 6, 7, 13, 12, 0},
  {1, 2, 9, 8, 13, 12},
  {1, 2, 10, 11, 0, 0}
};

// sens des aiguilles pour chaque itineraire :
// 0 = aiguille en position directe,
// 1 = aiguille deviee,
// >1 = aiguille non utilisee par l'itineraire
const byte ITI_aig[ITI_nb][AIG_nb] = {
  {0, 1, 255, 255, 0},    // Itineraire 1 : aiguilles 1 et 5 directes ; aiguille 2 deviee ; aiguilles 3 et 4 inutilisees
  {0, 0, 255, 1, 1},      // Sens des aiguilles pour l'itineraire 2
  {1, 255, 1, 0, 1},      // Sens des aiguilles pour l'itineraire 3
  {1, 255, 0, 255, 255}   // Etc...
};

/***************************************************************************
 * 
 * BOUTONS DE COMMANDE
 * 
 ***************************************************************************/
const byte CDE_nb = max(AIG_nb, ITI_nb);  // nombre boutons utilises = le plus grands entre aiguilles et itineraires
byte CDE_btn[CDE_nb];                     // ensemble des boutons de commande, commençant apres MODE_btn
byte CDE_etat[CDE_nb];                    // etat de ces boutons (HIGH = relaches, LOW = enfonces)


/***************************************************************************
 * 
 * TCO
 * 
 ***************************************************************************/

// initialisation de la suite de leds :
const int LUM = 50;     // luminosité générale des leds (0 à 255 max)
const int CDE = 12;     // broche qui commande la barre de leds
const int NB_leds = 14; // nombre de leds à commander
Adafruit_NeoPixel strip = Adafruit_NeoPixel(NB_leds, CDE, NEO_GRB + NEO_KHZ800);

// les couleurs utilisees pour les etats des aiguilles (directe = Vert, deviee = Rouge) :
const uint32_t AIG_couleur[2] = {strip.Color(0, 255, 0), strip.Color(255, 0, 0)};

// les couleurs utilisees pour les itineraires (Rouge, Ambre, Vert, Bleu) :
const uint32_t ITI_couleur[ITI_nb] = {strip.Color(255, 0, 0), strip.Color(255, 195, 0), strip.Color(0, 255, 0), strip.Color(0, 0, 255)};


/***************************************************************************
 * 
 * Fonctions
 * 
 ***************************************************************************/
//
// Animation lors du choix d'un itineraire :
//
void animItineraire(int numero, boolean cde) {
  // on éteint toutes les leds avant de commencer
  strip.clear();
  // on allume toutes les leds de l'itineraire les unes après les autres
  for (int led=0; led<ITI_led_max; led++) {
    if (ITI_led[numero][led] == 0) break;
    strip.setPixelColor(ITI_led[numero][led]-1, ITI_couleur[numero]);
    strip.show();   
    delay(50);
  }
  // et on pilote les aiguilles associees si on est en mode cde
  if (cde) {
    for (int aig=0; aig<AIG_nb; aig++) {
      if (ITI_aig[numero][aig] < 2) {           // si l'aiguille est a traiter = son etat est 0 (droite) ou 1 (deviee)
        AIG_etat[aig] = ITI_aig[numero][aig];   // on indique l'etat souhaite
        cdeAiguille(aig);                       // et on pilote l'aiguille
      }
    }
  }
}

//
// Animation lors du changement d'etat d'une aiguille :
//
void animAiguille(int numero) {
  // on eteind la led de l'ancien etat
  strip.setPixelColor(AIG_led[numero][!AIG_etat[numero]] - 1, 0);
  // et on allume celle du nouvel etat
  strip.setPixelColor(AIG_led[numero][AIG_etat[numero]] - 1, AIG_couleur[AIG_etat[numero]]);
  strip.show();
  // et on pilote l'aiguille
  cdeAiguille(numero);
}

//
// Animation de demarrage, permettant de tester toutes les leds :
//
void animDebut() {
  // défilement des 4 itineraires les uns après les autres
  for (int i=0; i<ITI_nb; i++) {
    animItineraire(i, false);
    delay(950);
  }
  // extinction de l'ensemble des LEDs en fin d'animation
  strip.clear();
  strip.show();
}

//
// Initialisation du mode aiguilles / itineraires
//
void initMode(boolean mode) {
  if (mode == LOW) {                      // mode itineraires
    digitalWrite(MODE_led, HIGH);         // on allume la led associee
    animItineraire(ITI_defaut - 1, true); // on allume les leds de l'itineraire par defaut en mode cde
  }
  else {                                // mode aiguilles
    digitalWrite(MODE_led, LOW);        // on eteind la led associee aux itineraires
    strip.clear();                      // on eteind toutes les leds
    for (int i=0; i<AIG_nb; i++) {
      AIG_etat[i] = AIG_defaut[i];      // on recupere l'etat par defaut de chaque aiguille
      animAiguille(i);                  // et on la traite
    }
  }
}

//
// Commande de changement d'etat d'une aiguille
//
void cdeAiguille(int numero) {
  AIG_broche = AIG_first + 2*numero + int(AIG_etat[numero]); // calcul de la broche a activer
  digitalWrite(AIG_broche, HIGH);                       // activation
  delay(ACTIVATION);                                    // duree de l'impulsion de commande
  digitalWrite(AIG_broche, LOW);                        // arret de l'activation
}

//
// Test des boutons avec anti-rebond logiciel :
//
boolean testBtn(int numero){
  CDE_etat[numero] = false;
  int etat = digitalRead(CDE_btn[numero]);
  int tempo = millis();
  while ((millis() - tempo) < DELAI_ANTI_REBOND) {}
  if (etat == LOW){CDE_etat[numero] = true;}
  return CDE_etat[numero];
}

/***************************************************************************
 * 
 * SETUP
 * 
 ***************************************************************************/

void setup() {

  // initialisation des generalites
  pinMode(MODE_btn, INPUT_PULLUP);    // bouton de choix de mode en entree avec resistance de rappel 5V
  pinMode(MODE_led, OUTPUT);          // led associee en sortie
  digitalWrite(MODE_led, LOW);        // eteinte par defaut

  // initialisation des broches pilotant les aiguilles
  for (int i=0; i<AIG_nb; i++) {
    AIG_broche = AIG_first + 2*i;       // calcul de la broche de debut de l'aiguille
    pinMode(AIG_broche, OUTPUT);        // broche de l'etat directe en sortie
    pinMode(AIG_broche + 1, OUTPUT);    // broche de l'etat deviee en sortie
    digitalWrite(AIG_broche, LOW);      // broche de l'etat directe au repos
    digitalWrite(AIG_broche + 1, LOW);  // broche de l'etat deviee au repos
  }

  // initialisation des boutons de commande
  for (int i=0; i<CDE_nb; i++) {
    CDE_btn[i] = MODE_btn + i + 1;      // broche du bouton, derrière la broche de MODE_btn
    pinMode(CDE_btn[i], INPUT_PULLUP);  // en entree avec resistance de rappel 5V
    CDE_etat[i] = HIGH;                 // par defaut, les boutons sont relaches
  }
  
  // initialisation des leds :
  strip.begin();            // démarrage
  strip.clear();            // toutes les leds éteintes
  strip.setBrightness(LUM); // réglage de la luminosité qui peut s'avérer trop forte par défaut
  strip.show();             // on envoie les données
  // animation de début :
  animDebut();

  // on est en mode aiguilles ou itineraires ? :
  MODE_etat == digitalRead(MODE_btn);   // lecture de l'etat initial du bouton de choix de mode
  initMode(MODE_etat);                  // initialisation associee
  
}


/***************************************************************************
 * 
 * BOUCLE LOOP
 * 
 ***************************************************************************/

void loop() {
  
  // A-t-on change de mode ?
  if (digitalRead(MODE_btn) != MODE_etat) {
    MODE_etat = !MODE_etat;         // on change de mode enregistre
    initMode(MODE_etat);            // et on initialise le nouveau mode
  }

  // Test des boutons en fonction du mode actuel :
  if (MODE_etat == LOW) {               // mode itineraires
    for (int numero = 0; numero < ITI_nb; numero++) {
      if (testBtn(numero)) {            // si le bouton est appuye
        animItineraire(numero, true);   // on active l'itineraire associe en mode cde
      }
    }
  }
  else {                                        // mode aiguilles
    for (int numero = 0; numero < AIG_nb; numero++) {
      if (testBtn(numero)) {                    // si le bouton est appuye
        AIG_etat[numero] = !AIG_etat[numero];   // on change l'etat de l'aiguille
        animAiguille(numero);                   // et on active le nouvel etat
      }
    }
  }
}
